// Head Tracking Device Code
// 
// This code was written under the supervision of Dr. Aman Behal at the University of Central Florida
// Nanoscience Technology Center.  By using this code, you agree that this software
// is distributed as-is, without warranty of any kind.
//
// This code was written using NaturalPoint's OptiTrack Software 
// and the Templatized C++ Command Line Parser (TCLAP).
//
// IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, CONSEQUENTIAL OR ANY OTHER
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
// GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THE
// PROGRAM, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. YOU
// HEREBY WAIVE ANY CLAIMS FOR DAMAGES OF ANY KIND AGAINST
// CONTRIBUTORS WHICH MAY RESULT FROM YOUR USE OF THE PROGRAM.
//

#define PI 3.14159265
#include "tclap/CmdLine.h"	//	If this is before windows.h it causes problems...why?  I don't know
#define _WIN32_WINNT 0x0501
#include <windows.h>
#include <math.h>
#include <string>
#include <iostream>
#include <algorithm>
#include <conio.h>    //== For sample simplicity: kbhit()

// Includes required for optitrack

#include <objbase.h>
#include <atlbase.h>
#include "optitrack.h"
#import  "optitrack.tlb"

//	Declare some namespaces in use

using namespace std;
using namespace TCLAP;

void LeftClick(bool ctrl_click = false);
void RightClick(bool ctrl_click = false);
void FindOurCamera(INPCameraCollection *collection, long count, CComPtr<INPCamera> &camera, long mod, long rev, long* rate);
void PrintCameraInfo(INPCamera *camera);
void TrackPoint(INPCamera *camera, int monResX, int monResY, bool ctrl_click, int train_timeout, int click_timeout);
void TrackVector(INPCamera *camera, int monResX, int monResY, int train_timeout, int click_timeout, bool ctrl_click, bool pauseClick, int click_strength, float sF);

/*
 @param	monResX	The display's width in a Windows environment.
 @param	monResY	The display's height in a Windows environment.
 */
int main(int argc, char** argv) {
	float monResX, monResY;
	long cameraCount = 0, rate = 0;

	//	Begin parsing the command line arguments
	try {
		CmdLine cmd("The head tracking device utilizes infrared sensing for motion detection. One or more reflectors can be deployed to track user motions. These reflectors can be easily attached to a baseball cap, for instance, allowing the user to move their head to navigate the graphical user interface in two dimensions, or command the robot in six dimenions.\nHead tracker is a project that interfaces with the NaturalPoint OptiTrack API, currently it only supports the TrackIR3 Pro series of cameras, and allows for a user to control the mouse with just the use of their head.", ' ', "1.0.0", true);

		//	Arguments for point mode
		SwitchArg pointMode("p", string("point"), string("Use the head tracker in point mode. [DEFAULT]"), false);

		//	Arguments for vector mode
		SwitchArg vectorMode("v", string("vector"), string("Use the head tracker in vector mode."), false);
		ValueArg<int> click_strength("s", string("clickStrength"), string("The strength at which the user must move their head left and right to generate a click event.  Only for Vector Mode."), false, 8, string("default=8"), cmd);
		ValueArg<float> scaling_factor("f", string("scalingFactor"), string("A number between 0 - 100.  Zero is less sensitive, and 100 is more sensative.  Only for Vector Mode."), false, 25, string("default=25"), cmd);
		SwitchArg pause_click("u", string("pauseClick"), string("If you want to have the pointer hold still after click event until next click event."), cmd, false);

		//	General settings for the application
		ValueArg<float> click_timeout("k", string("clickTimeout"), string("Time in seconds [<1] that a user has to hold on a pixel before sending allowing click event."), false, (float)1.4, string("default=1.4"), cmd);
		ValueArg<float> train_timeout("t", string("trainTimeout"), string("Time in seconds [<5] that will be used for training the head tracking algorithm."), false, 12.5, string("default=12.5"), cmd);
		SwitchArg quiet("q", string("quiet"), string("If you want to hide all of the detailed information about the camera."), cmd, false);
		SwitchArg ctrl_click("c", string("ctrlClick"), string("Should the CTRL button be pressed when the click signal is sent."), cmd, false);

		cmd.xorAdd(pointMode, vectorMode);	//	Allow either point mode or vector mode only
		cmd.parse(argc, argv);	//	Parse the command line arguments

		//	Get the monitor resolution
		monResX = (float)GetSystemMetrics(SM_CXSCREEN);
		monResY = (float)GetSystemMetrics(SM_CYSCREEN);

		//	Print the header
		cout << endl;
		cout.width(50);
		cout.fill('-');
		cout << "Automated Headtracker";
		cout.width(31);
		cout.fill('-');
		cout << " ";

		cout.width(44);
		cout.fill(' ');
		cout << "Monitor Resolution: ";
		printf("%0.fx%0.f\n\n", monResX, monResY);

		//== Initialize Microsoft COM Interop ================----
		CoInitialize(NULL);

		//== Initialize OptiTrack COM Component ==============----
		CComPtr<INPCameraCollection> cameraCollection;
		CComPtr<INPCamera> camera;

		cameraCollection.CoCreateInstance(CLSID_NPCameraCollection);

		printf("Enumerating Camera(s)...\r");

		//== Enumerate (Identify) Available Cameras ==========----
		cameraCollection->Enum();

		printf("Counting the Camera(s)...\r");

		//== Determine Available Cameras =====================----
		cameraCollection->get_Count(&cameraCount);

		if(cameraCount > 0) {	//	Check to see that camera(s) are attached to the system
			printf("% 2.i Camera(s) Where Detected...\r\n\n", cameraCount);

			//	Load our camera into memory
			FindOurCamera(cameraCollection, cameraCount, camera, NP_HW_MODEL_TRACKIR, NP_HW_REVISION_TRACKIR_PRO, &rate);

			if(camera != NULL) {	//	Make sure the camera was found
				if(!quiet.getValue())	//	Print information about the camera if the user didn't say not to
					PrintCameraInfo(camera);

				if(pointMode.getValue())	//	Track the point by the user's type
					TrackPoint(camera, (int)monResX, (int)monResY, ctrl_click.getValue(), (int)train_timeout.getValue()*rate, (int)click_timeout.getValue()*rate);
				else if(vectorMode.getValue())
					TrackVector(camera, (int)monResX, (int)monResY, (int)train_timeout.getValue()*rate, (int)click_timeout.getValue()*rate, ctrl_click.getValue(), pause_click.getValue(), click_strength.getValue(), scaling_factor.getValue());
			} else {
				cerr << "The camera that was searched for could not be found.  Please specify a different camera model." << endl;
			}


		} else {
			cerr << "No cameras where detected.  Please attach a camera and re-run the application.\r" << endl;
		}

		//== Always Clean-up COM Objects ================----
		camera.Release();
		cameraCollection.Release();

		//== Uninitialize Microsoft COM Interop =============----
		CoUninitialize();

		printf("\nApplication Exit.\n");
		system("PAUSE");
		return 0;

	} catch (ArgException &e) {
		cerr << "error: " << e.error() << " for arg " << e.argId() << endl;
	}
}

/*
 Function simulates the button click in a Windows environment
 this only will work in Windows XP/2003 or Vista.  Basically,
 all systems where SendInput has been added.
 */
void LeftClick(bool ctrl_click) {
  INPUT    Input={0};
  // left down
  Input.type      = INPUT_MOUSE;
  Input.mi.dwFlags  = MOUSEEVENTF_LEFTDOWN;
  ::SendInput(1,&Input,sizeof(INPUT));

  // left up
  ::ZeroMemory(&Input,sizeof(INPUT));
  Input.type      = INPUT_MOUSE;
  Input.mi.dwFlags  = MOUSEEVENTF_LEFTUP;
  ::SendInput(1,&Input,sizeof(INPUT));

  if(ctrl_click) {
	  keybd_event(VK_CONTROL,0x9d,0 , 0); // Ctrl Press
	  keybd_event(VK_CONTROL,0x9d,KEYEVENTF_KEYUP,0); // Ctrl Release
  }
}

/*
 Function simulates the button click in a Windows environment
 this only will work in Windows XP/2003 or Vista.  Basically,
 all systems where SendInput has been added.
 */
void RightClick(bool ctrl_click) {
  INPUT    Input={0};
  // right down
  Input.type      = INPUT_MOUSE;
  Input.mi.dwFlags  = MOUSEEVENTF_RIGHTDOWN;
  ::SendInput(1,&Input,sizeof(INPUT));

  // right up
  ::ZeroMemory(&Input,sizeof(INPUT));
  Input.type      = INPUT_MOUSE;
  Input.mi.dwFlags  = MOUSEEVENTF_RIGHTUP;
  ::SendInput(1,&Input,sizeof(INPUT));

  if(ctrl_click) {
	  keybd_event(VK_CONTROL,0x9d,0 , 0); // Ctrl Press
	  keybd_event(VK_CONTROL,0x9d,KEYEVENTF_KEYUP,0); // Ctrl Release
  }
}

/*
 Function is passed the reference to a camera object.  It then
 takes and prints relavant information about that camera on the
 screen.
 */
void PrintCameraInfo(INPCamera *camera) {
	long serial, width, height, model, revision, rate;
	camera->get_SerialNumber(&serial);
	camera->get_Width       (&width);
	camera->get_Height      (&height);
	camera->get_Model       (&model);
	camera->get_Revision    (&revision);
	camera->get_FrameRate   (&rate);

	printf(" Camera %d\n",serial);
	printf(" -------------------------\n");
	printf(" Resolution: %dx%d\n",width,height);
	printf(" Revision  : 0x%8x\n",revision);
	printf(" Model     : 0x%8x\n",model);
	printf(" Frame rate: %d\n" ,rate);

	if(model == NP_HW_MODEL_TRACKIR)
		printf(" TrackIR2 or TrackIR3\n");
	if(revision == NP_HW_REVISION_TRACKIR_PRO)
		printf(" TrackIR Pro\n");

	cout << endl;
}

/*
 Find our camera in the collection, using the camera referenced object
 to store the pointer to the object.
 */
void FindOurCamera(INPCameraCollection *collection, long count, CComPtr<INPCamera> &camera, long mod, long rev, long* rate) {
	long model, revision;
	for(int ii = 0; ii < count; ii++) {
		collection->Item(0, &camera);		// Load the camera into memory

		camera->get_Model(&model);	//	Load model info
		camera->get_Revision(&revision);	//	Load revision num
		camera->get_FrameRate(rate);

		if(model == mod && revision == rev)
			ii = count + 1;
		else
			camera.Release();	// Always Clean-up COM Objects
	}
}

/*
 Track a point on the TrackIR camera.  Use this tracked point 
 to force the cursor on screen to move with the user's input motions.
 */
void TrackPoint(INPCamera *camera, int monResX, int monResY, bool ctrl_click, int train_timeout, int click_timeout) {
	int maxX = 0, maxY = 0, minY = INT_MAX, minX = INT_MAX, frameCount = 0, prevX = 0, prevY = 0, dwelled = 0, timeToDwell = 99, timeToTrain = 500;
	bool moved = false;

	if(click_timeout > timeToDwell) timeToDwell = click_timeout;
	if(train_timeout > timeToTrain) timeToTrain = train_timeout;

	//	Print a header
	printf("% 16.15s % 20.14s % 13.9s % 11.9s\n", "TrackIR Obj Pos", "Calc Mouse Pos", "Max [X,Y]", "Min [X,Y]");
	cout.width(64);
	cout.fill('-');
	cout << " " << endl;

	//	Create required come objects
	CComPtr<INPCameraFrame> frame;
	CComPtr<INPSmoothing> smoothingDot;

	smoothingDot.CoCreateInstance(CLSID_NPSmoothing);

	//	Perform tracking
	camera->Open();
	camera->Start();
    while(!_kbhit()) {
		camera->GetFrame(0, &frame);

		if(frame != NULL) {
			long lCount = 0;
			frame->get_Count(&lCount);

			for (int i = 0; i < lCount; i++) {
				CComPtr<INPObject> spObject;

				frame->Item(i, &spObject);

				long lRank;

				spObject->get_Rank(&lRank);
				if (lRank == 1) {
					// this is the currently tracked object...
					CComVariant varX, varY, varArea;

					// values are double's so we need to use a variant to query for them
					spObject->get_X(&varX);
					spObject->get_Y(&varY);
					spObject->get_Area(&varArea);

					// smooth the data
					smoothingDot->Update(varX, varY);

					// grab the smoothed data
					smoothingDot->get_X(&varX);
					smoothingDot->get_Y(&varY);

					//	Check if movement occurred
					if( (abs(prevX - (int)varX.dblVal) > 1) ) moved = true;
					else moved = false;
					
					if(!moved) {	//	Only check to see if it moved in the y direction if X direction didn't change
						if ( (abs(prevY - (int)varY.dblVal) > 1)) moved = true;	//	Outside range
						else moved = false;
					}

					if(moved) {	//	If it moved then store the current X & Y position in the previous position vars
						prevX = (int)varX.dblVal;
						prevY = (int)varY.dblVal;
					}

					if(frameCount <= timeToTrain) {	//	For the first timeToTrain frames calibrate
						if(varX.dblVal > maxX) maxX = (int)varX.dblVal;
						if(varY.dblVal > maxY) maxY = (int)varY.dblVal;
						if(varX.dblVal < minX) minX = (int)varX.dblVal;
						if(varY.dblVal < minY) minY = (int)varY.dblVal;
						if(maxX-minX == 0) maxX++;
						if(maxY-minY == 0) maxY++;
						frameCount++;
					}

					varX.dblVal -= minX;
					varY.dblVal -= minY;

					varX.dblVal = varX.dblVal;
					varY.dblVal = varY.dblVal;

					int myWidth = maxX - minX;
					int myHeight = maxY - minY;
					
					float calcX = (float)(((monResX/myWidth)*(myWidth - varX.dblVal) >= 0) ? (monResX/myWidth)*(myWidth - varX.dblVal) : 0);
					float calcY = (float)(((monResY/myHeight)*varY.dblVal >= 0) ? (monResY/myHeight)*varY.dblVal : 0);

					// update the UI
					printf("[%07.i,%07.i] | [%08.3f,%08.3f] | [%03.i,%03.i] | [%03.i,%03.i] | [%03.i]\r", (int)varX.dblVal, (int)varY.dblVal, calcX, calcY, maxX, maxY, minX, maxY, dwelled);

					SetCursorPos((int)calcX, (int)calcY);
					
					if(!moved) {
						dwelled++;
						if(dwelled >= timeToDwell) {
							LeftClick(ctrl_click);	//	Pass cmd line argument to perform ctrl click
							dwelled = 0;
						}
					} else {
						dwelled = 0;
					}
				}	//	End if checking for rank
			}	//	End for loop for objects
			frame->Free();
			frame.Release();	//	Always Clean-up COM Objects
		}	// End check for frame that is not null
	}	//	End while statement

	camera->Stop();
	camera->Close();
	smoothingDot.Release();	//	Always Clean-up COM Objects
}

/*
 Track a vector on the TrackIR camera.  Use this tracked point 
 to force the cursor on screen to move with the user's input motions.
 */
void TrackVector(INPCamera *camera, int monResX, int monResY, int train_timeout, int click_timeout, bool ctrl_click, bool pauseClick, int click_strength, float sF) {
	CComVariant varYaw, varPitch, varRoll, varX, varY, varZ, varEmpty;
	char status = ' ';
	int frameCount = 0, timeToTrain = 500, prevX = 0, prevY = 0, dwelled = 0, timeToDwell = 99;
	double off_yaw = 0, off_pitch = 0, off_roll = 0, off_xx = 0, off_yy = 0, off_zz = 0, max_yaw = 0, min_yaw = 0, max_pitch = 0, min_pitch = 0;
	double xx = 0, yy = 0, zz = 0, yaw = 0, pitch = 0, roll = 0, k = sF*((double)1/(double)10), deltaT = (double)1/(double)120;
	bool left = false, right = false, reset = false, stop = false, moved = false, up = true;

	if(click_timeout > timeToDwell) timeToDwell = click_timeout;
	if(train_timeout > timeToTrain) timeToTrain = train_timeout;

	CComPtr<INPCameraFrame> frame;
    CComPtr<INPVector> vector;

	vector.CoCreateInstance(CLSID_NPVector);

	//	Print a header
	printf("% 22.15s % 12.3s % 11.5s % 9.4s\n", "TrackIR [X,Y,Z]", "Yaw", "Pitch", "Roll");
	cout.width(60);
	cout.fill('-');
	cout << " " << endl;

	//	Perform tracking
	camera->Open();
	camera->Start();
    while(!stop) {
		if(_kbhit()) {
			int pressed_key = _getch();	//	Get key pressed
			if(pressed_key == 114)
				reset = true;
			else if(pressed_key == 27)
				stop = true;
			else if(pressed_key == 224)
				reset = true;
		}

		camera->GetFrame(0, &frame);

		if(frame != NULL) {
			// update the vector positions
			vector->Update(camera, frame);

			// get the values back
			vector->get_Yaw(&varYaw);
			vector->get_Pitch(&varPitch);
			vector->get_Roll(&varRoll);
			vector->get_X(&varX);
			vector->get_Y(&varY);
			vector->get_Z(&varZ);

			//	Check if movement occurred
			if( (abs(prevX - (int)varX.dblVal) > 1) ) moved = true;
			else moved = false;
			
			if(!moved) {	//	Only check to see if it moved in the y direction if X direction didn't change
				if ( (abs(prevY - (int)varY.dblVal) > 1)) moved = true;	//	Outside range
				else moved = false;
			}

			if(moved) {	//	If it moved then store the current X & Y position in the previous position vars
				prevX = (int)varX.dblVal;
				prevY = (int)varY.dblVal;
			}

			if(frameCount <= timeToTrain) {	//	For the first timeToTrain frames calibrate
				if(varYaw.dblVal > max_yaw) max_yaw = varYaw.dblVal;
				if(varYaw.dblVal < min_yaw) min_yaw = varYaw.dblVal;
				if(varPitch.dblVal > max_pitch) max_pitch = varPitch.dblVal;
				if(varPitch.dblVal < min_pitch) min_pitch = varPitch.dblVal;
				if(max_yaw-min_yaw == 0) max_yaw++;
				if(max_pitch-min_pitch == 0) max_pitch++;
				frameCount++;
			}

			if(reset) {	// If the reset command was given set the reset value
				vector->Reset();
				off_yaw = -varYaw.dblVal;
				off_pitch = -varPitch.dblVal;
				off_roll = -varRoll.dblVal;
				reset = false;		//	reset the reset condition
			}

			varYaw.dblVal += off_yaw;
			varPitch.dblVal += off_pitch;
			varRoll.dblVal += off_roll;

			yaw = yaw - k*yaw*deltaT + k*varYaw.dblVal*deltaT;	//	Filter signal
			yaw = ((int)(yaw*100))/((double)100);	//	Truncate to 2 decimal locations
			pitch = pitch - k*pitch*deltaT + k*varPitch.dblVal*deltaT;	//	Filter signal
			pitch = ((int)(pitch*100))/((double)100);	//	Truncate to 2 decimal locations
			roll = roll - k*roll*deltaT + k*varRoll.dblVal*deltaT;	//	Filter signal
			roll = ((int)(roll*100))/((double)100);	//	Truncate to 2 decimal locations

			zz = zz - k*zz*deltaT + k*varZ.dblVal*deltaT;
			zz = ((int)(zz*100))/((double)100);	//	Truncate to 2 decimal locations

			if(frameCount <= timeToTrain) {	//	While in the training period
				xx = zz*tan((yaw*PI)/180) + (zz*tan((yaw*PI)/180)*(sF/100)) + (monResX/2);
				yy = -zz*tan((pitch*PI)/180) + (-zz*tan((pitch*PI)/180)*(sF/100)) + (monResY/2);
			} else {
				if(yaw < 0) {	//	If yaw is negative use min yaw
					xx = (-monResX / (2*tan((min_yaw*PI)/180))) * tan((yaw*PI)/180) + (monResX/2);
				} else if(yaw >= 0) {	//	If yaw is positive use max yaw
					xx = (monResX / (2*tan((max_yaw*PI)/180))) * tan((yaw*PI)/180) + (monResX/2);
				}

				//	Now that the system has been calibrated use the max and min pitches to determine the x and y coords
				if(pitch < 0) {	//	If pitch is negative use min pitch
					yy = (monResY / (2*tan((min_pitch*PI)/180))) * tan((pitch*PI)/180) + (monResY/2);
				} else if(pitch >= 0) {	//	If pitch is positive use max pitch
					yy = (-monResY / (2*tan((max_pitch*PI)/180))) * tan((pitch*PI)/180) + (monResY/2);
				}
			}

			// update the UI
			if(!(frameCount <= timeToTrain)) {
				if(!left && varRoll.dblVal > click_strength) {
					status = 'l';
					LeftClick(ctrl_click);
					up = !up;
					left = true;
				} else if(left && varRoll.dblVal <= click_strength) {
					left = false;
					status = ' ';
				} else if(!right && varRoll.dblVal < (-1*click_strength)) {
					status = 'r';
					RightClick(ctrl_click);
					up = !up;
					right = true;
				} else if(right && varRoll.dblVal > (-1*click_strength)) {
					right = false;
					status = ' ';
				}
			}

			printf("[%08.3f,%08.3f,%06.2f] | [%06.2f] | [%06.2f] | [%06.2f] | [%c] | [%c] \r", varX.dblVal, varY.dblVal, zz, yaw, pitch, roll, status, ((frameCount <= timeToTrain) ? 't' : ' '));
			if(!pauseClick || up) SetCursorPos((int)xx, (int)yy);	//	Only allow movement when up

			if(!moved) {
				dwelled++;
				if(dwelled >= timeToDwell) {
					//LeftClick(ctrl_click);	//	Pass cmd line argument to perform ctrl click
					dwelled = 0;
				}
			} else {
				dwelled = 0;
			}

			frame->Free();
			frame.Release();	//	Always Clean-up COM Objects
		}	// End if frame NULL statement
	}	//	End while loop

	//	Always Clean-up COM Objects
	camera->Stop();
	camera->Close();
}
